
/////////////////////////////////////////////////////////////////////
// Author: Husni Che Ngah
// Email:  hcnhusni@streamyx.com
/////////////////////////////////////////////////////////////////////

#pragma once

#include <math.h>
#include <xutility>
using namespace std;

// CHCNObject command target

class CHCN_Object : public CObject
{
public:
	typedef enum HCN_OBJECTTYPE			// Structured enumeration with type defined
	{
		HCN_None=0x00,
		HCN_Line,
		HCN_Polygon,
		HCN_Rectangle,
		HCN_Ellipse,
		HCN_Text,
		HCN_Composite
	};
	typedef union HCN_VALUE { INT nVal; UINT uVal; float fVal; double dwVal; };
	typedef union HCN_COLOR { INT nColor; COLORREF cColor; double dwColor; };

	CHCN_Object();
	virtual ~CHCN_Object();

protected:
	HCN_OBJECTTYPE	m_obType;
	
public:
	BOOL m_bSelected;

	virtual void Serialize(CArchive& ar);

	void SetObjectType(HCN_OBJECTTYPE obType) { m_obType = obType; }; 
	HCN_OBJECTTYPE GetObjectType() { return m_obType; };

	//////////////////////////////////////////////////////////////////////////////////////
	Rect Point2Rect(Point &pt1, Point &pt2) const
	{
		Rect r;
		r.X = _MIN(pt1.X, pt2.X);
		r.Y = _MIN(pt1.Y, pt2.Y);
		r.Width = _MAX(pt1.X, pt2.X) - _MIN(pt1.X, pt2.X);
		r.Height = _MAX(pt1.Y, pt2.Y) - _MIN(pt1.Y, pt2.Y);
		return r;
	};

	CRect CPoint2CRect(CPoint &cpt1, CPoint &cpt2) const
	{
		CRect r;
		r.left = _MIN(cpt1.x, cpt2.x);
		r.top = _MIN(cpt1.y, cpt2.y);
		r.right = r.left + (_MAX(cpt1.x, cpt2.x) - _MIN(cpt1.x, cpt2.x));
		r.bottom = r.top + (_MAX(cpt1.y, cpt2.y) - _MIN(cpt1.y, cpt2.y));
		return r;
	};

	///////////////////////////////////////////////////////////////////////////////////////
	Rect GetRectCentered(Point &ptCenter, Point &ptEnd) const
	{
		Rect r;
		int nDX = _MAX(ptCenter.X, ptEnd.X) - _MIN(ptCenter.X, ptEnd.X);
		int nDY = _MAX(ptCenter.Y, ptEnd.Y) - _MIN(ptCenter.Y, ptEnd.Y);
		r.X = ptCenter.X - nDX;
		r.Y = ptCenter.Y - nDY;
		r.Width = nDX * 2;
		r.Height = nDY * 2;
		return r;
	};

	CRect GetRectCentered(CPoint &cptCenter, CPoint &cptEnd) const
	{
		CRect r;
		int nDX = _MAX(cptCenter.x, cptEnd.x) - _MIN(cptCenter.x, cptEnd.x);
		int nDY = _MAX(cptCenter.y, cptEnd.y) - _MIN(cptCenter.y, cptEnd.y);
		r.left = cptCenter.x - nDX;
		r.top = cptCenter.y - nDY;
		r.right = r.left + (nDX * 2);
		r.bottom = r.top + (nDY * 2);
		return r;
	}

	////////////////////////////////////////////////////////////////////////////////////
	Rect GetSquareCentered(Point &ptCenter, Point &ptEnd) const
	{
		Rect r;
		int nDX = _MAX(ptCenter.X, ptEnd.X) - _MIN(ptCenter.X, ptEnd.X);
		int nDY = _MAX(ptCenter.Y, ptEnd.Y) - _MIN(ptCenter.Y, ptEnd.Y);
		int nMaxDelta = _MAX(nDX, nDY);
		r.X = ptCenter.X - nMaxDelta;
		r.Y = ptCenter.Y - nMaxDelta;
		r.Width = r.Height = nMaxDelta * 2;
		return r;
	};

	CRect GetSquareCentered(CPoint &cptCenter, CPoint &cptEnd) const
	{
		CRect r;
		int nDX = _MAX(cptCenter.x, cptEnd.x) - _MIN(cptCenter.x, cptEnd.x);
		int nDY = _MAX(cptCenter.y, cptEnd.y) - _MIN(cptCenter.y, cptEnd.y);
		int nMaxDelta = _MAX(nDX, nDY);
		r.left = cptCenter.x - nMaxDelta;
		r.top = cptCenter.y - nMaxDelta;
		r.right = r.left + (nMaxDelta * 2);
		r.bottom = r.top + (nMaxDelta * 2);
		return r;
	}

	///////////////////////////////////////////////////////////////////////////////////
	Rect GetCircleCentered(Point &ptCenter, Point &ptEnd) const
	{
		Rect r;
		double dwDX = _MAX(ptCenter.X, ptEnd.X) - _MIN(ptCenter.X, ptEnd.X);
		double dwDY = _MAX(ptCenter.Y, ptEnd.Y) - _MIN(ptCenter.Y, ptEnd.Y);
		double dwRad = sqrt((dwDX*dwDX) + (dwDY*dwDY));

		r.X = ptCenter.X - (UINT)dwRad;
		r.Y = ptCenter.Y - (UINT)dwRad;
		r.Width = r.Height = (UINT)dwRad * 2;
		return r;
	};

	CRect GetCircleCentered(CPoint &cptCenter, CPoint &cptEnd) const
	{
		CRect r;
		double dwDX = _MAX(cptCenter.x, cptEnd.x) - _MIN(cptCenter.x, cptEnd.x);
		double dwDY = _MAX(cptCenter.y, cptEnd.y) - _MIN(cptCenter.y, cptEnd.y);
		double dwRad = sqrt((dwDX*dwDX) + (dwDY*dwDY));

		r.left = cptCenter.x - (UINT)dwRad;
		r.top = cptCenter.y - (UINT)dwRad;
		r.right = r.left + ((UINT)dwRad * 2);
		r.bottom = r.top + ((UINT)dwRad * 2);
		return r;
	}

	Rect GetEllipseCentered(Point &ptCenter, Point &ptFirst, Point &ptSecond) const
	{
		Rect r;
		int nDX1 = _MAX(ptCenter.X, ptFirst.X) - _MIN(ptCenter.X, ptFirst.X); VERIFY(nDX1>0);
		int nDY1 = _MAX(ptCenter.Y, ptFirst.Y) - _MIN(ptCenter.Y, ptFirst.Y); VERIFY(nDY1>0);
		int nMaxDelta1 = _MAX(nDX1, nDY1);

		int nDX2 = _MAX(ptCenter.X, ptSecond.X) - _MIN(ptCenter.X, ptSecond.X); VERIFY(nDX2>0);
		int nDY2 = _MAX(ptCenter.Y, ptSecond.Y) - _MIN(ptCenter.Y, ptSecond.Y); VERIFY(nDY2>0);
		int nMaxDelta2 = _MAX(nDX2, nDY2);

		r.X = ptCenter.X - nMaxDelta1;
		r.Y = ptCenter.Y - nMaxDelta2;
		r.Width = nMaxDelta1 * 2;
		r.Height = nMaxDelta2 * 2;
		return r;
	};

	CRect GetEllipseCentered(CPoint &cptCenter, CPoint &cptFirst, CPoint &cptSecond) const
	{
		CRect r;
		int nDX1 = _MAX(cptCenter.x, cptFirst.x) - _MIN(cptCenter.x, cptFirst.x); VERIFY(nDX1>0);
		int nDY1 = _MAX(cptCenter.y, cptFirst.y) - _MIN(cptCenter.y, cptFirst.y); VERIFY(nDY1>0);
		int nMaxDelta1 = _MAX(nDX1, nDY1);

		int nDX2 = _MAX(cptCenter.x, cptSecond.x) - _MIN(cptCenter.x, cptSecond.x); VERIFY(nDX2>0);
		int nDY2 = _MAX(cptCenter.y, cptSecond.y) - _MIN(cptCenter.y, cptSecond.y); VERIFY(nDY2>0);
		int nMaxDelta2 = _MAX(nDX2, nDY2);
		
		r.left = cptCenter.x - nMaxDelta1;
		r.top = cptCenter.y - nMaxDelta2;
		r.right = r.left + (nMaxDelta1 * 2);
		r.bottom = r.top + (nMaxDelta2 * 2);
		return r;
	}

protected:
	DECLARE_SERIAL(CHCN_Object)
};


class CHCN_ObjectArray : public CObArray
{
public:
	CHCN_ObjectArray();
	virtual ~CHCN_ObjectArray();

protected:
	CHCN_Object::HCN_VALUE	m_hCurIndex;
	CHCN_Object::HCN_OBJECTTYPE m_hCurType;

public:
	void SetCurrentIndex(int nIndex) { m_hCurIndex.nVal=nIndex; };
	void SetCurrentIndex(double dwIndex) { m_hCurIndex.dwVal=dwIndex; };

	int GetCurrentIndex() { return m_hCurIndex.nVal; };
	
	CHCN_Object::HCN_OBJECTTYPE GetCurrentType() 
	{ 
		if(this->GetSize()>0)
		{
			CHCN_Object* pBase = GetObjectBaseClass(m_hCurIndex.nVal);
			m_hCurType = pBase->GetObjectType();
		}
		else
			m_hCurType=CHCN_Object::HCN_None;
		
		return m_hCurType; 
	};

	CHCN_Object* GetObjectBaseClass(int nIndex)
	{
		m_hCurIndex.nVal = nIndex;
		return (reinterpret_cast<CHCN_Object*>(GetAt(nIndex)));
	}

protected:
	DECLARE_SERIAL(CHCN_ObjectArray)
public:
	virtual void Serialize(CArchive& ar);
};

class CHCN_ObjectPoint : public CHCN_Object
{
public:
	CHCN_ObjectPoint();
	virtual ~CHCN_ObjectPoint();

	CHCN_ObjectPoint(const CHCN_ObjectPoint &p) 
	{
		m_dwX=p.m_dwX;
		m_dwY=p.m_dwY;
		SetObjectType(HCN_None);
	}

	CHCN_ObjectPoint operator=(const CHCN_ObjectPoint& p) 
	{
		m_dwX=p.m_dwX;
		m_dwY=p.m_dwY;
		SetObjectType(HCN_None);
		return *this;
	}

	CHCN_ObjectPoint(int nX, int nY) 
	{
		CHCN_Object::HCN_VALUE hX, hY;
		hX.nVal = nX;
		hY.nVal = nY;
		m_dwX = hX.dwVal;
		m_dwY = hY.dwVal;
		SetObjectType(HCN_None);
	}

	CHCN_ObjectPoint(double dwX, double dwY) 
	{
		m_dwX = dwX;
		m_dwY = dwY;
		SetObjectType(HCN_None);
	}

	CHCN_ObjectPoint(CPoint &p) 
	{
		CHCN_Object::HCN_VALUE hX, hY;
		hX.nVal = p.x;
		hY.nVal = p.y;
		m_dwX = hX.dwVal;
		m_dwY = hY.dwVal;
		SetObjectType(HCN_None);
	}

	CHCN_ObjectPoint(Point &p) 
	{
		CHCN_Object::HCN_VALUE hX, hY;
		hX.nVal = p.X;
		hY.nVal = p.Y;
		m_dwX = hX.dwVal;
		m_dwY = hY.dwVal;
		SetObjectType(HCN_None);
	}

	CHCN_ObjectPoint(PointF &pF) 
	{
		m_dwX = (double)pF.X;
		m_dwY = (double)pF.Y;
		SetObjectType(HCN_None);
	}

	CPoint GetPoint() const 
	{
		CHCN_Object::HCN_VALUE hX, hY;
		hX.dwVal = m_dwX;
		hY.dwVal = m_dwY;
		return CPoint(hX.nVal, hY.nVal);
	}

	Point GetGDIPoint() const 
	{
		CHCN_Object::HCN_VALUE hX, hY;
		hX.dwVal = m_dwX;
		hY.dwVal = m_dwY;
		return Point(hX.nVal, hY.nVal);
	}

	PointF GetGDIPointF() const 
	{
		return PointF((float)m_dwX, (float)m_dwY);
	}

	virtual void Serialize(CArchive &archive)
	{
		if(archive.IsStoring())
		{
			// Store data
			archive << m_dwX;
			archive << m_dwY;
		}
		else
		{
			// Load data;
			archive >> m_dwX;
			archive >> m_dwY;
		}
	};

	CHCN_Object* GetObjectBaseClass() { return (reinterpret_cast<CHCN_Object*>(this)); };

protected:
	double m_dwX;
	double m_dwY;

protected:
	DECLARE_SERIAL(CHCN_ObjectPoint)
};

class CHCN_ObjectLine : public CHCN_Object
{
public:
	CHCN_ObjectLine();
	virtual ~CHCN_ObjectLine();

	// Smart Constructors
	CHCN_ObjectLine(Point* points, INT_PTR nCount, Color &color, REAL rWidth);
	CHCN_ObjectLine(CHCN_ObjectArray* ptArray, Color &color, REAL rWidth);

	// Attribute inquiry
	REAL GetLineWidth() const { return m_lineWidth; };
	Color GetLineColor() const { return m_lineColor; };
	Point GetPointAt(INT_PTR nIdx) const { CHCN_ObjectPoint* pt = (CHCN_ObjectPoint*)m_ptArray->GetAt(nIdx); return pt->GetGDIPoint(); };
	CPoint GetCPointAt(INT_PTR nIdx) const { CHCN_ObjectPoint* pt = (CHCN_ObjectPoint*)m_ptArray->GetAt(nIdx); return pt->GetPoint(); };
	INT_PTR GetPointCount() const { return m_ptArray->GetSize(); };
	Point* GetPointArray() const;

	//
	void SetPointAt(Point &point, INT_PTR nIdx) {CHCN_ObjectPoint* pt = new CHCN_ObjectPoint(point); m_ptArray->SetAt(nIdx, pt); };
	void SetPointAt(CPoint &point, INT_PTR nIdx) {CHCN_ObjectPoint* pt = new CHCN_ObjectPoint(point); m_ptArray->SetAt(nIdx, pt); };
	void SetPoints(CHCN_ObjectArray* ptArray) { m_ptArray = ptArray; };
	void SetPoints(Point* points, INT_PTR nCnt);
	
	//
	void SetLineWidth(REAL width) { m_lineWidth = width; };
	void SetLineColor(Color &color) { m_lineColor = color; };
	
	// Class intepreter
	CHCN_Object* GetObjectBaseClass() { return (reinterpret_cast<CHCN_Object*>(this)); };

	// Attributes
protected:
	Color				m_lineColor;
	REAL				m_lineWidth;
	CObArray*			m_ptArray;			// Store series of points

public:
	BOOL				m_bCurved;
	BOOL				m_bShowNode;

	void DrawObject(Graphics& graphics);	// Graphics function
	void DrawNormalLine(Graphics& graphics);	// Graphics function
	void DrawCurvedLine(Graphics& graphics);	// Graphics function
	void DrawSegmentNode(Graphics& graphics);	// Graphics function

	virtual void Serialize(CArchive& ar);

protected:
	DECLARE_SERIAL(CHCN_ObjectLine)
public:
	BOOL IsSegmentHit(CPoint& pt);
	BOOL IsNodeHit(CPoint& pt);
};


